import itertools

import networkx as nx
import numba
import numpy as np
import pandas as pd
import pygeohash_fast
from joblib import Parallel, delayed

from gpsmatcher.emission import emission_matrix
from gpsmatcher.gps import process_gps
from gpsmatcher.graph import process_graph
from gpsmatcher.precomputation_emission import process_dic_cand_edges
from gpsmatcher.transition import transition_matrix


def get_predecessor(G, edge, edge_pre):
    if edge[0] == edge_pre[1]:
        return(edge_pre)
    else:
        path = nx.shortest_path(G, edge_pre[1], edge[0])
        return((path[-2], path[-1]))

def get_sucessor(G, edge, edge_suc):
    if edge[1] == edge_suc[0]:
        return(edge_suc)
    else:
        path = nx.shortest_path(G, edge[1], edge_suc[0])
        return((path[0], path[1]))

def get_new_edge_id(cand_edgeid, emit_p, edges):
    nb_rows, nb_cols = np.shape(cand_edgeid)
    new_edgeid = np.zeros((nb_rows,))
    for i in range(nb_rows):
        val_max = -1
        for j in range(nb_cols):
            edgeid = cand_edgeid[i, j]
            if edgeid != -1:
                val = emit_p[i, edgeid]
                if val > val_max:
                    val_max = val
                    edgeid_min = edgeid
        if val_max != -1:
            new_edgeid[i] = edgeid_min
        else:
            new_edgeid[i] = edges[i]
    return(new_edgeid)

def correct_edge(graph, states, state, sub_edges, id2edges, edges2id, emit_p):
    edge2state = dict(zip(sub_edges, states))
    edges = [id2edges[sub_edges[int(i)]] for i in state]
    edges_no_redundancy = [i[0] for i in itertools.groupby(edges)]
    edges_no_redundancy_count = [len(list(i[1])) for i in itertools.groupby(edges)]

    pre_edges = [-1]  + edges_no_redundancy[0:-1]
    next_edges = edges_no_redundancy[1:] + [-1] 
    sucessors = [get_sucessor(graph, edges_no_redundancy[i], next_edges[i]) for i in range(len(edges_no_redundancy) - 1)] + [-1]
    predecessors = [-1] + [get_predecessor(graph, edges_no_redundancy[i], pre_edges[i]) for i in range(1, len(edges_no_redundancy))]  
    cand_edges = [[sucessors[i], edges_no_redundancy[i], predecessors[i]] for i in range(len(sucessors) )] 
    cand_edgeid = [[edge2state[edges2id[i]] if i!=-1 else -1 for i in sub_list ] for sub_list in cand_edges]
    cand_edgeid = np.array(cand_edgeid)
    cand_edgeid = np.repeat(cand_edgeid, edges_no_redundancy_count, axis=0)
    new_edgeid = get_new_edge_id(cand_edgeid, emit_p, edges)
    return(new_edgeid)


def get_cand_edge_gps(gps, cand_edge):
    """
    Filter candidate edges based on GPS data.

    Parameters
    ----------
    gps : pandas.DataFrame
        GPS data.

    cand_edge : dict
        Dictionary mapping geohashes to candidate edges.

    Returns
    -------
    cand_edge : dict
        Updated dictionary with filtered candidate edges.
    """
    all_geohashes = set(pygeohash_fast.encode_many(gps['lon'].values, gps['lat'].values, 8))
    cand_edge = {k: cand_edge[k] for k in cand_edge.keys() & all_geohashes}
    return(cand_edge)

def format_result(gps, gps_mm):
    """
    Format map-matching results.

    Parameters
    ----------
    gps : pandas.DataFrame
        GPS data.

    gps_mm : pandas.DataFrame
        Map-matched GPS data.

    Returns
    -------
    gps : pandas.DataFrame
        Updated GPS data with edge information.

    gps_mm : pandas.DataFrame
        Formatted map-matching results.
    """
    gps.reset_index(drop=True, inplace=True)
    gps_mm['shortest_path_nodes'] = gps_mm.apply(lambda row: row['map_match'][1], axis=1)
    gps_mm['edges'] = gps_mm.apply(lambda row: row['map_match'][0], axis=1)
    gps = gps[['lon', 'lat', 'ID_trip']]
    gps['edge'] = gps_mm['edges'].explode().reset_index(drop=True)
    #gps_mm = gps_mm["shortest_path_nodes"].reset_index()
    return(gps, gps_mm)

def mm_precomputation(G, beta=1/500, radius=150, save=True, folder_name="mm_input", show_print=True):
    """
    Perform map-matching precomputation.

    Parameters
    ----------
    G : networkx.DiGraph
        The road network graph.

    beta : float, optional
        Parameter for transition matrix computation.

    radius : int, optional
        Radius for candidate edges computation.

    save : bool, optional
        Whether to save the precomputed data.

    folder_name : str, optional
        Folder name for saving data.

    show_print : bool, optional
        Whether to print progress messages.

    Returns
    -------
    G_mm : networkx.DiGraph
        Processed road network graph for map-matching.

    trans : scipy.sparse.csr_matrix
        Transition matrix for map-matching.

    dic_candidates : dict
        Dictionary mapping geohashes to candidate edges.

    id2edges : dict
        Dictionary mapping edge_id to edge tuples.
    """
    G_mm, id2edges, edges2id = process_graph(G, max_length=500, save=save, folder_name=folder_name, show_print=show_print)
    trans = transition_matrix(G_mm, 120, beta=beta, save=save, folder_name=folder_name, show_print=show_print)
    dic_candidates = process_dic_cand_edges(G_mm, radius = radius, save=save, folder_name=folder_name, show_print=show_print)
    return(G_mm, trans, dic_candidates, id2edges, edges2id)

def mm_gps(gps, G_mm, trans, dic_candidates, id2edges, edges2id, alpha=0.1, radius=150):
    """
    Given the precomputation, map-match gps data (without multiprocessing)

    Parameters
    ----------
    gps : pandas.DataFrame
        GPS data.

    G_mm : networkx.DiGraph
        Processed road network graph for map-matching.

    trans : scipy.sparse.csr_matrix
        Transition matrix for map-matching.

    dic_candidates : dict
        Dictionary mapping geohashes to candidate edges.

    id2edges : dict
        Dictionary mapping edge_id to edge.

    alpha : float, optional
        Parameter for emission matrix computation.

    radius : int, optional
        Radius for candidate edges computation.

    Returns
    -------
    gps : pandas.DataFrame
        Updated GPS data with edge information.

    gps_mm : pandas.DataFrame
        Map-matched GPS data. Each ID_trip with most likely path in the graph.
    """
    gps, gps_mm, dic_geohash, dic_candidates = process_gps(gps, dic_candidates)
    emit = emission_matrix(gps, G_mm, dic_geohash, dic_candidates, alpha = alpha, radius = radius)
    gps_mm['map_match'] = gps_mm.apply(lambda row: one_traj_mm(row['traj'], G_mm, trans, emit, id2edges, edges2id, row['sub_edges'], row['first_emission'], row['last_emission']), axis=1)
    gps, gps_mm = format_result(gps, gps_mm)
    return(G_mm, gps, gps_mm)

def mm_gps_parrallel(gps, G_mm, trans, dic_candidates, id2edges, edges2id, alpha=0.1, radius=150, nb_cores=4):
    """
    Given the precomputation, map-match gps data (with multiprocessing)

    Parameters
    ----------
    gps : pandas.DataFrame
        GPS data.

    G_mm : networkx.DiGraph
        Processed road network graph for map-matching.

    trans : scipy.sparse.csr_matrix
        Transition matrix for map-matching.

    dic_candidates : dict
        Dictionary mapping geohashes to candidate edges.

    id2edges : dict
        Dictionary mapping edge_id to edge.

    alpha : float, optional
        Parameter for emission matrix computation.

    radius : int, optional
        Radius for candidate edges computation.

    nb_cores : int, optional
        Number of CPU cores to use.

    Returns
    -------
    gps : pandas.DataFrame
        Updated GPS data with edge information.

    gps_mm : pandas.DataFrame
        Map-matched GPS data. Each ID_trip with most likely path in the graph.
    """
    dic_candidates = get_cand_edge_gps(gps, dic_candidates)
    all_chunks_gps =  np.array_split(gps, nb_cores)
    result = Parallel(n_jobs=nb_cores, verbose=20)(delayed(mm_gps)(chunk_gps, G_mm, trans, dic_candidates, id2edges, edges2id, alpha=alpha, radius=radius) for chunk_gps in all_chunks_gps)
    gps = pd.concat([i_result[0] for i_result in result])
    gps_mm = pd.concat([i_result[1] for i_result in result])
    return(G_mm, gps, gps_mm)

def big_gps_file_mm(G, gps, radius = 150, alpha = 0.1, beta=1/500, nb_rows_chunk = 100000, save=True, folder_name="mm_input", show_print=True, parrallel = False, nb_cores=3):
    """
    Perform map-matching on large GPS data where the data is splitted into chunks to process.

    Parameters
    ----------
    G : networkx.DiGraph
        The road network graph.

    gps : pandas.DataFrame
        GPS data with three columns lon, lat, ID_trip.

    save : bool, optional
        Whether to save the precomputation or load if already computed.

    radius : int, optional
        Radius for candidate edges computation.

    alpha : float, optional
        Parameter for emission matrix computation.

    beta : float, optional
        Parameter for transition matrix computation.

    folder_name : str, optional
        Folder name for saving data.

    show_print : bool, optional
        Whether to print progress messages.

    parallel : bool, optional
        Whether to use parallel processing.

    nb_cores : int, optional
        Number of CPU cores to use.

    Returns
    -------
    gps : pandas.DataFrame
        Updated GPS data with edge information.

    gps_mm : pandas.DataFrame
        Map-matched GPS data. Each ID_trip with most likely path in the graph.
    """
    G_mm, trans, dic_candidates, id2edges, edges2id = mm_precomputation(G, beta=beta, radius=radius, save=save, folder_name=folder_name, show_print=show_print)
    
    nb_chunks = len(gps) // nb_rows_chunk
    all_chunks_gps =  np.array_split(gps, nb_chunks)
    all_edges  = []
    all_gps_mm  = []
    for idx, chunk_gps in enumerate(all_chunks_gps):
        print("{}/{} chunk done".format(idx+1, nb_chunks))
        if parrallel:
            _, chunk_gps, chunk_gps_mm = mm_gps_parrallel(chunk_gps, G_mm, trans, dic_candidates, id2edges, edges2id, alpha=alpha, radius=radius, nb_cores=nb_cores)
        else:
            _, chunk_gps, chunk_gps_mm = mm_gps(chunk_gps, G_mm, trans, dic_candidates, id2edges, edges2id, alpha=alpha, radius=radius)
        all_edges.extend(chunk_gps['edge'].values)
        all_gps_mm.append(chunk_gps_mm)
    
    all_gps_mm = pd.concat(all_gps_mm)
    gps['edge'] = all_edges
    return(G_mm, gps, all_gps_mm)

def gps_file_mm(G, gps, save=True, radius = 150, alpha = 0.1, beta=1/500, folder_name="mm_input", show_print=True, parrallel = False, nb_cores=3):
    """
    Perform map-matching on GPS data.

    Parameters
    ----------
    G : networkx.DiGraph
        The road network graph.

    gps : pandas.DataFrame
        GPS data with three columns lon, lat, ID_trip.

    save : bool, optional
        Whether to save the precomputation or load if already computed.

    radius : int, optional
        Radius for candidate edges computation.

    alpha : float, optional
        Parameter for emission matrix computation.

    beta : float, optional
        Parameter for transition matrix computation.

    folder_name : str, optional
        Folder name for saving data.

    show_print : bool, optional
        Whether to print progress messages.

    parallel : bool, optional
        Whether to use parallel processing.

    nb_cores : int, optional
        Number of CPU cores to use.

    Returns
    -------
    gps : pandas.DataFrame
        Updated GPS data with edge information.

    gps_mm : pandas.DataFrame
        Map-matched GPS data. Each ID_trip with most likely path in the graph
    """
    G_mm, trans, dic_candidates, id2edges, edges2id = mm_precomputation(G, beta=beta, radius=radius, save=save, folder_name=folder_name, show_print=show_print)
    if parrallel:
        G_mm, gps, gps_mm = mm_gps_parrallel(gps, G_mm, trans, dic_candidates, id2edges, edges2id, alpha=alpha, radius=radius, nb_cores=nb_cores)
    else:
        G_mm, gps, gps_mm = mm_gps(gps, G_mm, trans, dic_candidates, id2edges, edges2id, alpha=alpha, radius=radius)
    return(G_mm, gps, gps_mm)

def one_traj_mm(GPS_traj, graph, transition_matrix, emission_matrix, id2edges, edges2id,sub_edges, start, end):
    """
    Perform map-matching on a single GPS trajectory.

    Parameters
    ----------
    GPS_traj : numpy.ndarray
        Array containing GPS observations.

    graph : networkx.DiGraph
        The road network graph.

    transition_matrix : scipy.sparse.csr_matrix
        Transition matrix for the graph.

    emission_matrix : scipy.sparse.csr_matrix
        Emission matrix for the graph.

    id2edges : dict
        Dictionary mapping edge_id to edge of the graph.

    sub_edges : list
        List of edge IDs candidates considered for map-matching.

    start : int
        Start index in emission_matrix.

    end : int
        End index in emission_matrix).

    Returns
    -------
    edges : list
        List of edges matched to the GPS trajectory.

    path : list
        List of nodes representing the matched path on the road network graph.
    """
    try:
        emit_p = (emission_matrix[start:end+1, : ][:, sub_edges].toarray())/100
        trans_p = (transition_matrix[sub_edges, :][:, sub_edges].toarray())/100
        start_p = np.ones(len(sub_edges))/len(sub_edges)
        obs = np.array([i for i in range(len(GPS_traj))])
        states = np.array([i for i in range(len(sub_edges))])
        (V,prob, state,path) = fast_viterbi(obs, states, start_p, trans_p, emit_p)
        if prob != 0:
            edges = [id2edges[sub_edges[int(i)]] for i in state]
            new_edges = correct_edge(graph, states, state, sub_edges,id2edges, edges2id,emit_p)
            new_edges = [id2edges[sub_edges[int(i)]] for i in new_edges]
            edges_without_redundancy = [i[0] for i in itertools.groupby(new_edges)]
            #edges_without_redundancy = [i[0] for i in itertools.groupby(edges)] #be careful
            path = [nx.shortest_path(graph, edges_without_redundancy[i][1], edges_without_redundancy[i+1][0]) for i in range(len(edges_without_redundancy) - 1)]
            path = [edges_without_redundancy[0][0]] + [item for sublist in path for item in sublist] + [edges_without_redundancy[-1][1]]
            return(new_edges, path)
        else:
            return([np.nan]*len(GPS_traj),[])
    except:
        return(["Problem"]*len(GPS_traj), ["Problem"])

@numba.jit(nopython=True)
def fast_viterbi(obs, states, start_p, trans_p, emit_p):
    """
    Implement the Viterbi algorithm for map-matching.

    Parameters
    ----------
    obs : numpy.ndarray
        Array containing observation indices.

    states : numpy.ndarray
        Array containing state indices.

    start_p : numpy.ndarray
        Initial state probabilities.

    trans_p : numpy.ndarray
        Transition probabilities.

    emit_p : numpy.ndarray
        Emission probabilities.

    Returns
    -------
    V : numpy.ndarray
        Viterbi matrix.

    prob : float
        Probability of the most likely path.

    seq_nodes : numpy.ndarray
        Sequence of node indices representing the most likely path.

    path : numpy.ndarray
        most likely path matrix.
    """
    nb_obs = obs.shape[0]
    nb_states = states.shape[0]

    V = np.zeros((nb_obs, nb_states))
    path = np.zeros((1,nb_states))

    # Initialize base cases (t == 0)
    for idx, y in enumerate(states):
        V[0, y] = start_p[y] * emit_p[ obs[0],y]
        path[0, idx] = y

    for itime in range(1, nb_obs):   
        states_cand = np.where(emit_p[ obs[itime],:] != 0)[0]
        newpath = np.zeros((itime + 1 ,states_cand.shape[0]))
        states_bis =np.where(V[itime-1, :] != 0)[0]
        for idx, y in enumerate(states_cand):
            prob = 0
            for y0 in states_bis:
                if V[itime-1, y0] * trans_p[y0,y] * emit_p[ obs[itime],y] > prob:
                    prob = V[itime-1, y0] * trans_p[y0,y] * emit_p[ obs[itime],y]
                    state = y0
            V[itime, y] = prob
            if prob>0: #check this if problem
                row = np.where(path[-1,:] == state)[0][0]
                newpath[:itime, idx] = path[:, row]
                newpath[-1, idx] = y
        # Don't need to remember the old paths
        path = newpath

    idx_max = np.argmax(V[itime,:])
    (prob, state) = V[itime, idx_max], idx_max
    if prob !=0:
        row = np.where(path[-1,:] == state)[0][0]
        seq_nodes = path[:, row]
    else:
        path = np.zeros((1,nb_states))
        seq_nodes = np.zeros((1,nb_states))[:,0]
    return (V,prob, seq_nodes,path)


import os

import networkx as nx
import numpy as np
import pandas as pd
try:
    import pygeohash as pgh
except ImportError:
    raise ImportError(
        "The 'pygeohash' package is required for this module but is not installed.\n"
        "Please install it directly from GitHub:\n"
        "pip install 'git+https://github.com/wdm0006/pygeohash.git@a89e6e794f0dde4e066048ec16c867e945a680c1'"
    )
if not hasattr(pgh, 'get_adjacent'):
    raise ImportError(
        "The installed version of 'pygeohash' does not have the required 'get_adjacent' method.\n"
        "Please install the correct version directly from GitHub:\n"
        "pip install 'git+https://github.com/wdm0006/pygeohash.git@a89e6e794f0dde4e066048ec16c867e945a680c1'"
    )
import pygeohash_fast
from gpsmatcher.utils import load_param, save_param


def process_dic_cand_edges(G, radius = 150, save=True, folder_name="mm_input", show_print=True):
    """
    Process a dictionary of candidate edges for geohashes in a graph: load if dic_cand_edges has been already computed else dic_cand_edges is computed.

    Parameters
    ----------
    G : networkx.DiGraph
        The input graph.

    radius : float, optional
        Radius for computing candidate edges. Default is 150 meters.

    save : bool, optional
        Whether to save the processed candidate edges. Default is True.

    folder_name : str, optional
        Folder name for saving/loading. Default is "mm_input".

    show_print : bool, optional
        Whether to print a message indicating the completion of the operation. Default is True.

    Returns
    -------
    dic_cand_edges : dict
        Dictionary mapping geohashes to candidate edges.
    """
    
    if save and (os.path.isfile(folder_name + '/cand_edges.pickle')):
        dic_cand_edges = load_param(folder_name, "cand_edges", show_print=show_print)

    else:
        dic_cand_edges = get_dic_cand_edges(G, radius = radius, show_print=show_print)

        if save and (~os.path.isfile(folder_name + '/cand_edges.pickle')):
            save_param(folder_name, "cand_edges", dic_cand_edges)
    return(dic_cand_edges)


def get_all_neigbors(hash_central, dir_geo, dic_neighbor, nb_steps = 19):
    """
    Get all neighbors in a specified direction for a given geohash.

    Parameters
    ----------
    hash_central : str
        Central geohash.

    dir_geo : str
        Direction of neighbors ("top", "bottom", "left", "right").

    dic_neighbor : dict
        Dictionary to store computed neighbors.

    nb_steps : int, optional
        Number of steps in each direction. Default is 19.

    Returns
    -------
    all_neigbors_dir : list
        List of all geohashes in the specified direction.
    """
    all_keys = dic_neighbor.keys()
    all_neigbors_dir = [hash_central]
    i = 0
    for i in range(nb_steps):
        if (hash_central, dir_geo) in all_keys:
            neigbhor_dir = dic_neighbor[(hash_central, dir_geo)]
        else:
            neigbhor_dir = pgh.get_adjacent(hash_central, dir_geo)
            dic_neighbor[(hash_central, dir_geo)] = neigbhor_dir
            
            all_neigbors_dir.append(neigbhor_dir)
        hash_central = neigbhor_dir
    return(all_neigbors_dir, dic_neighbor)

def compute_neighbors(hash_central, nb_steps_lon, nb_steps_lat, dic_neighbor):
    """
    Compute neighbors for a central geohash in both horizontal and vertical directions.

    Parameters
    ----------
    hash_central : str
        Central geohash.

    nb_steps_lon : int
        Number of steps in the horizontal direction.

    nb_steps_lat : int
        Number of steps in the vertical direction.

    dic_neighbor : dict
        Dictionary to store computed neighbors.

    Returns
    -------
    all_geohashes : list
        List of all computed geohashes.
    """
    
    all_neigbors_dir_north, dic_neighbor = get_all_neigbors(hash_central, "top",dic_neighbor, nb_steps=nb_steps_lat)
    all_neigbors_dir_south, dic_neighbor = get_all_neigbors(hash_central, "bottom", dic_neighbor, nb_steps=nb_steps_lat)
    vertical_geoshes = all_neigbors_dir_south + all_neigbors_dir_north
    all_geohashes = vertical_geoshes.copy()
    for hash_v in vertical_geoshes:
        all_neigbors_dir_west, dic_neighbor = get_all_neigbors(hash_v, "left", dic_neighbor, nb_steps=nb_steps_lon)
        all_neigbors_dir_est, dic_neighbor  = get_all_neigbors(hash_v, "right", dic_neighbor, nb_steps=nb_steps_lon)
        all_geohashes.extend(all_neigbors_dir_west + all_neigbors_dir_est)
    
    return(all_geohashes, dic_neighbor)

def get_all_emission_tiles(geohashes, edge, nb_steps_lon, nb_steps_lat, dic_neighbor, dic_cand_edges): 
    """
    Get all geohashes neighbors for a list of central geohashes (geohashes of all edges).

    Parameters
    ----------
    geohashes : list
        List of central geohashes.

    edge : int
        Edge ID associated with the geohashes.

    nb_steps_lon : int
        Number of steps in the horizontal direction.

    nb_steps_lat : int
        Number of steps in the vertical direction.

    dic_neighbor : dict
        Dictionary to store computed neighbors.

    dic_cand_edges : dict
        Dictionary mapping geohashes to candidate edges.

    Returns
    -------
    all_geohashes_tiles : list
        List of all geohashes for a list of central geohashes (geohashes of all edges).
    """
    all_geohashes_tiles = []
    dic_neighbor = {}
    for i in geohashes:
        sub_geohashes, dic_neighbor = compute_neighbors(i, nb_steps_lon, nb_steps_lat ,dic_neighbor)
        all_geohashes_tiles.extend(sub_geohashes)
    all_geohashes_tiles = list(set(all_geohashes_tiles))
    
    keys_geohash = dic_cand_edges.keys()
    for i in all_geohashes_tiles:
        if i not in keys_geohash:
            dic_cand_edges[i] = [edge]
        else:
            dic_cand_edges[i] += [edge]
    return(all_geohashes_tiles)

def get_number_tiles(radius):
    """
    Get the number of steps in the horizontal and vertical directions for geohashes within a radius.

    Parameters
    ----------
    radius : float
        Radius for computing emission tiles.

    Returns
    -------
    nb_lat_steps : int
        Number of steps in the vertical direction.

    nb_lon_steps : int
        Number of steps in the horizontal direction.
    """
    width, height = 38.2, 19.1
    nb_lon_steps = int((radius // width) + 1)
    nb_lat_steps = int((radius // height) + 1)
    return(nb_lat_steps, nb_lon_steps)

def get_dic_cand_edges(G, radius = 150, show_print=True):
    """
    Get a dictionary of candidate edges for geohashes in a graph.

    Parameters
    ----------
    G : networkx.DiGraph
        The input graph.

    radius : float, optional
        Radius for computing candidate edges. Default is 150 meters.

    show_print : bool, optional
        Whether to print a message indicating the completion of the operation. Default is True.

    Returns
    -------
    dic_cand_edges : dict
        Dictionary mapping geohashes to candidate edges.
    """
    edges2id =nx.get_edge_attributes(G, 'edge_id')
    edge2geohash = nx.get_edge_attributes(G, 'geohashes')
    edge_geohash = pd.DataFrame(edge2geohash.items(),columns=['edge', 'geohash'])
    edge_geohash['edge_id'] = edge_geohash['edge'].map(edges2id)
    dic_cand_edges = {}
    dic_neighbor = {}
    nb_lat_steps, nb_lon_steps = get_number_tiles(radius)
    edge_geohash['geohash_bis'] = edge_geohash.apply(lambda row : get_all_emission_tiles(row['geohash'], row['edge_id'], nb_lon_steps, nb_lat_steps, dic_neighbor, dic_cand_edges), axis=1)
    if show_print:
        print("Get candidate edges for all geohashes done")
    return(dic_cand_edges)



import os

from scipy.sparse import csr_matrix

os.environ["USE_PYGEOS"] = "1"
import geopandas as gpd
import networkx as nx
import numpy as np
import pandas as pd
import pickle

from gpsmatcher.utils import load_param, print_step, save_param


def chunks_dist_computation(chunk_gps, cand_edges, radius=150, alpha=0.1):
    """
    Compute distances between GPS points in a chunk and candidate edges.

    Parameters
    ----------
    chunk_gps : geopandas.GeoDataFrame
        GPS points in the current chunk.

    cand_edges : pandas.DataFrame
        Candidate edges with associated geometries.

    radius : float, optional
        Maximum distance for a GPS point to be matched to an edge. Default is 150 meters.

    alpha : float, optional
        Tuning parameter for the distance calculation. Default is 0.1.

    Returns
    -------
    dist_df : pandas.DataFrame
        DataFrame with columns 'id', 'edge', and 'dist' representing GPS points and distance to the candidate edges.
    """
    
    dist_df = chunk_gps[["id", "geometry"]].join(cand_edges, how='inner')
    dist_df.reset_index(drop=True, inplace=True)
    dist_df['dist'] = dist_df['geometry'].distance(dist_df['geometry_edge'])
    dist_df.drop(columns=['geometry', 'geometry_edge'], inplace=True)
    dist_df = dist_df[dist_df['dist'] < radius]
    dist_df['dist'] =  ((np.exp(-0.5* ((dist_df['dist'].values / 1000)/alpha)**2)/(np.sqrt(2*np.pi) * alpha))*100).astype(np.int16)
    return(dist_df[['id','edge','dist']])


def emission_matrix(gps, G, dic_geohash, dic_candidates, alpha = 0.1, radius = 150, chunk=True, nb_chunks = 20, show_print=True):  
    """
    Generate an emission matrix for map-matching based on GPS data and candidate edges.

    Parameters
    ----------
    gps : geopandas.GeoDataFrame
        GPS points with 'id' and 'geometry' columns.

    G : networkx.Graph
        Road network represented as a graph.

    dic_geohash : dict
        Dictionary mapping geohashes to integers.

    dic_candidates : dict
        Dictionary mapping geohashes to candidate edges.

    alpha : float, optional
        Tuning parameter for the distance calculation. Default is 0.1.

    radius : float, optional
        Maximum distance for a GPS point to be matched to an edge. Default is 150 meters.

    chunk : bool, optional
        Whether to process GPS data in chunks. Default is True.

    nb_chunks : int, optional
        Number of chunks to split the GPS data into if 'chunk' is True. Default is 20.

    Returns
    -------
    emission_matrix : scipy.sparse.csr_matrix
        Emission matrix representing the likelihood of GPS points emitting from edges.
    """

    #with open('debug_emmision.pickle', 'wb') as handle:
    #    pickle.dump((gps, G, dic_geohash, dic_candidates), handle, protocol=pickle.HIGHEST_PROTOCOL)
    if show_print:
        print_step("Start process emission")
    geom_df = (nx.to_pandas_edgelist(G)).rename(columns={'edge_id': 'edge'})
    geom_df = gpd.GeoDataFrame(geom_df[['edge', 'geometry']], geometry=geom_df['geometry'], crs=4326)
    geom_df.to_crs(3035, inplace=True)
    print(dic_candidates)
    cand_edges = pd.DataFrame.from_dict(dic_candidates.items())
    cand_edges.columns=['geohash', 'edge']
    cand_edges['geohash_int'] = cand_edges['geohash'].map(dic_geohash)
    cand_edges.drop(columns=["geohash"], inplace=True)
    cand_edges = cand_edges.explode("edge")
    cand_edges.reset_index(drop=True, inplace=True)
    cand_edges = cand_edges.astype(np.uint32)
    cand_edges = cand_edges.merge(geom_df, on='edge')
    cand_edges.set_index('geohash_int', inplace=True)
    cand_edges.rename(columns={'geometry': 'geometry_edge'}, inplace=True)
    if chunk:
        gps_splited = np.array_split(gps, nb_chunks)
        dist_df = []
        for idx, chunk_gps in enumerate(gps_splited):
            dist_df.append(chunks_dist_computation(chunk_gps, cand_edges, radius=radius, alpha=alpha))
        dist_df = pd.concat(dist_df)

    else:
        dist_df = gps[["id", "geometry"]].join(cand_edges, how='inner')
        dist_df.reset_index(drop=True, inplace=True)
        dist_df['dist'] = dist_df['geometry'].distance(dist_df['geometry_edge'])
        dist_df.drop(columns=['geometry', 'geometry_edge'], inplace=True)
        dist_df = dist_df[dist_df['dist'] < radius]
        dist_df['dist'] =  ((np.exp(-0.5* ((dist_df['dist'].values / 1000)/alpha)**2)/(np.sqrt(2*np.pi) * alpha))*100).astype(np.int16)
        print(dist_df)
    nb_rows, nb_cols = int(dist_df['id'].max() + 1), len(G.edges())
    emission_matrix = csr_matrix((dist_df['dist'].values, ( dist_df['id'].values, dist_df['edge'].values)), shape=(nb_rows,nb_cols))
    return(emission_matrix)


import os

import geopandas as gpd
import networkx as nx
import numpy as np
import pandas as pd
import pygeohash_fast
from shapely.geometry import LineString, MultiPoint, Point
from shapely.ops import snap, split

from gpsmatcher.utils import (load_param, print_step, remove_unneccesarry_att,
                              save_param)


def compute_edge_length(G, show_print=True):
    """
    Compute and assign lengths to edges in a networkx graph based on their geometries.

    Parameters
    ----------
    G : networkx.DiGraph
        The graph with edges containing 'geometry' attributes.

    show_print : bool, optional
        Whether to print a message indicating the completion of the computation. Default is True.

    Returns
    -------
    G : networkx.DiGraph
        The input graph with added 'length' attributes for each edge.
    """
    
    edges2geom = nx.get_edge_attributes(G, 'geometry')
    geom_edges = pd.DataFrame(edges2geom.items(),columns=['edge', 'geometry'])
    geom_edges['points'] = geom_edges.apply(lambda x: [y for y in x['geometry'].coords], axis=1)
    geom_points = geom_edges.explode('points')
    geom_points[['lon', 'lat']] = pd.DataFrame(geom_points['points'].tolist(), index=geom_points.index)
    geom_points['dist'] = haversine(geom_points["lat"], geom_points["lon"], geom_points["lat"].shift(-1), geom_points["lon"].shift(-1))
    geom_points['dist'] = 1000 * geom_points['dist'] * (geom_points['edge'] == geom_points['edge'].shift(-1))
    edges2length = geom_points.groupby('edge')['dist'].sum().to_dict()
    nx.set_edge_attributes(G, edges2length, 'length')
    
    if show_print:
        print("Compute edge legnth done")
    return(G)


def haversine(lat1, lon1, lat2, lon2, to_radians=True, earth_radius=6371):
    """
    Calculate the great circle distance between two points on the Earth.

    Parameters
    ----------
    lat1, lon1, lat2, lon2 : float
        Latitude and longitude coordinates of two points.

    to_radians : bool, optional
        Whether the input coordinates are in radians. Default is True.

    earth_radius : float, optional
        Earth's radius in kilometers. Default is 6371.

    Returns
    -------
    distance : float
        The great circle distance between the two points.
    """
    if to_radians:
        lat1, lon1, lat2, lon2 = np.radians([lat1, lon1, lat2, lon2])

    a = np.sin((lat2-lat1)/2.0)**2 + \
        np.cos(lat1) * np.cos(lat2) * np.sin((lon2-lon1)/2.0)**2
    return earth_radius * 2 * np.arcsin(np.sqrt(a))

def geom2length(geom_edge):
    """
    Calculate the length of a LineString geometry.

    Parameters
    ----------
    geom_edge : shapely.geometry.LineString
        LineString geometry.

    Returns
    -------
    length : float
        Length of the LineString in meters.
    """
    coords = np.array(geom_edge.coords)
    length = 1000 *sum(haversine(coords[:-1, 1], coords[:-1, 0], coords[1:, 1], coords[1:, 0]))
    return(length)

def interpolate_graph(G, max_length = 500, show_print=True):
    """
    Interpolate long edges in a graph by adding new nodes and edges.

    Parameters
    ----------
    G : networkx.DiGraph
        The input graph.

    max_length : float, optional
        Maximum length for an edge. If an edge's length exceeds this value, it will be interpolated.
        Default is 500 meters.

    show_print : bool, optional
        Whether to print a message indicating the completion of the interpolation. Default is True.

    Returns
    -------
    G : networkx.DiGraph
        The graph with interpolated edges.
    """
    edges2length =nx.get_edge_attributes(G, 'length')
    long_edges = [key for key, value in edges2length.items() if value >= max_length]
    start_node_id = max(G.nodes) + 1
    for edge in long_edges:
        atts_edges = G.edges[edge]
        atts_nodes = {key: value for (key, value) in G.nodes[edge[0]].items()}
        nb_new_nodes = int(atts_edges['length'] // max_length)
        line = atts_edges['geometry']
        new_nodes = MultiPoint([Point(line.coords[0])] + [line.interpolate(i/(nb_new_nodes + 1), normalized=True) for i in range(1,nb_new_nodes+1)] + [Point(line.coords[-1])])
        int_nodes = new_nodes.geoms[1:-1].geoms
        for j in range(len(int_nodes)):
            atts_nodes['x'] = int_nodes[j].x
            atts_nodes['y'] = int_nodes[j].y
            atts_nodes['edge'] = edge
            G.add_node(start_node_id + j, **atts_nodes)

        new_edges = split(snap(line, new_nodes, 1), new_nodes).geoms
        parent_edge_length, parent_edge_weight = atts_edges['length'], atts_edges['weight']
        
        for i in range(len(new_edges)):
            atts_edges['geometry'] = new_edges[i]
            atts_edges['length'] = geom2length(atts_edges['geometry'])
            atts_edges['weight'] = parent_edge_weight * atts_edges['length'] / parent_edge_length
            atts_edges['parent_edges'] = edge
            if i == 0:
                from_node = edge[0]
                to_node = start_node_id
            elif i == len(new_edges) - 1:
                from_node = start_node_id
                to_node = edge[1]
            else:
                from_node = start_node_id
                start_node_id += 1
                to_node = start_node_id
            G.add_edge(from_node, to_node, **atts_edges)
        start_node_id += 1
        G.remove_edge(edge[0], edge[1])
        
    if show_print:
        print("interpolate graph - interpolation distance: {} m - done".format(max_length))
    return(G)

def get_listgeohahses_edge(G, show_print=True):
    """
    Get geohashes associated with each edge in a graph.

    Parameters
    ----------
    G : networkx.DiGraph
        The input graph.

    show_print : bool, optional
        Whether to print a message indicating the completion of the operation. Default is True.

    Returns
    -------
    G : networkx.Graph
        The graph with added 'geohashes' attribute for each edge.
    """
    edges2interpolation = interpolation_edge_geom(G, dist_max = 5)
    geom_interpolation = pd.DataFrame(edges2interpolation.items(),columns=['edge', 'geometry'])
    geom_interpolation = gpd.GeoDataFrame(geom_interpolation, geometry=geom_interpolation['geometry'])
    geom_interpolation = geom_interpolation.explode('geometry', index_parts=True)
    geom_interpolation["lon"] = geom_interpolation["geometry"].x
    geom_interpolation["lat"] = geom_interpolation["geometry"].y
    geom_interpolation['geohash']   = pygeohash_fast.encode_many(geom_interpolation['lon'].values, geom_interpolation['lat'].values, 8)
    edge_geohash = geom_interpolation.groupby('edge')['geohash'].apply(set).apply(list).to_dict()
    nx.set_edge_attributes(G, edge_geohash, 'geohashes')
    
    if show_print:
        print("Get geoashes of edges done")
    return(G)

def edge_subsampling(edge_geom, edge_length, dist_max = 5):
    """
    Subsample points along an edge geometry to achieve a specified maximum distance between points.

    Parameters
    ----------
    edge_geom : shapely.geometry.LineString
        LineString geometry representing an edge.

    edge_length : float
        Length of the edge.

    dist_max : float, optional
        Maximum distance between subsampled points. Default is 5 meters.

    Returns
    -------
    geom_edge : shapely.geometry.MultiPoint
        Subsampled points along the edge geometry.
    """
    geom_edge, len_edge = edge_geom, round(edge_length) 
    if len_edge > dist_max:
        geom_edge = MultiPoint([geom_edge.interpolate(i/len_edge, normalized=True) for i in range(0,len_edge+1,dist_max)])
    elif len_edge ==0:
        geom_edge = MultiPoint(list(geom_edge.coords))
    else:
        geom_edge = geom_edge.boundary
    return(geom_edge)

def interpolation_edge_geom(G, dist_max = 5):
    """
    Interpolate edge geometries in a graph by subsampling points to achieve a specified maximum distance.

    Parameters
    ----------
    G : networkx.DiGraph
        The input graph.

    dist_max : float, optional
        Maximum distance between subsampled points. Default is 5 meters.

    Returns
    -------
    edge2newinterpolatedgeom : dict
        Dictionary mapping edges to their interpolated geometries.
    """
    interpolated_geom = [edge_subsampling(G.edges[i]['geometry'],G.edges[i]['length'], dist_max = dist_max) for i in G.edges]
    keys = list(G.edges)
    edge2newinterpolatedgeom = dict(zip(keys, interpolated_geom))
    return(edge2newinterpolatedgeom)

def fill_missing_geometry(G, show_print=True):
    """
    Fill missing 'geometry' attributes in edges of a graph by creating a LineString between nodes when missing.

    Parameters
    ----------
    G : networkx.DiGraph
        The input graph.

    show_print : bool, optional
        Whether to print a message indicating the completion of the operation. Default is True.

    Returns
    -------
    G : networkx.Graph
        The graph with filled 'geometry' attributes when missing.
    """
    for u, v in G.edges:
        atts = G.edges[u,v].keys()
        if 'geometry' not in atts:
            u_node = G.nodes[u]
            v_node = G.nodes[v]
            edge_geom = LineString((Point((u_node['x'], u_node['y'])), Point((v_node['x'], v_node['y']))))
            G.edges[u,v]['geometry'] = edge_geom
            
    if show_print:
        print("Fill missing geometry done")
    return(G)



def process_graph(G, max_length = 500, save=True, folder_name="mm_input", show_print=True):
    """
    Process a graph by removing unnecessary attributes, filling missing geometries, computing edge lengths,
    interpolating long edges, assigning geohashes to edges, and optionally saving the processed graph.

    Parameters
    ----------
    G : networkx.DiGraph
        The input graph.

    max_length : float, optional
        Maximum length for an edge. If an edge's length exceeds this value, it will be interpolated.
        Default is 500 meters.

    save : bool, optional
        Whether to save the processed graph. Default is True.

    folder_name : str, optional
        Folder name for saving/loading. Default is "mm_input".

    show_print : bool, optional
        Whether to print progress messages. Default is True.

    Returns
    -------
    G : networkx.DiGraph
        The processed graph.

    id2edges : dict
        Dictionary mapping orignal edge to a new edge_id.
    """
    if show_print:
        print_step("Start the process graph")
        
    if save and (os.path.isfile(folder_name + '/graph.pickle')):
        G = load_param(folder_name, "graph", show_print=show_print)
        edges = nx.to_pandas_edgelist(G)
        edges['edge'] = list(zip(edges["source"], edges["target"]))
        id2edges = edges.set_index("edge_id")['edge'].to_dict()
        edges2id = edges.set_index("edge")['edge_id'].to_dict()
    
    else:
        G = remove_unneccesarry_att(G)
        G = fill_missing_geometry(G, show_print=show_print)
        G = compute_edge_length(G, show_print=show_print)
        G = interpolate_graph(G, max_length = max_length, show_print=show_print)
        
        edges = nx.to_pandas_edgelist(G)
        edges['edge'] = list(zip(edges["source"], edges["target"]))
        edges['edge_id'] = range(len(edges))
        edges2id = edges.set_index("edge")['edge_id'].to_dict()
        id2edges = edges.set_index("edge_id")['edge'].to_dict()
        nx.set_edge_attributes(G, edges2id, 'edge_id')
        G = get_listgeohahses_edge(G, show_print=show_print)

        if save:
            save_param(folder_name, "graph", G)
    return(G, id2edges, edges2id)


import os

os.environ["USE_PYGEOS"] = "0"
import geopandas as gpd
import numpy as np
import pandas as pd
import pygeohash_fast


def process_gps(gps, cand_edge):
    """
    Process GPS data by encoding geohashes, mapping edges, and creating a GeoDataFrame.

    Parameters
    ----------
    gps : pandas.DataFrame
        DataFrame containing GPS data with columns 'lon', 'lat'.

    cand_edge : dict
        Dictionary mapping geohashes to candidate edges.

    Returns
    -------
    gps : gpd.GeoDataFrame
        GeoDataFrame with additional columns 'id', 'geohash_int', and 'geometry'.
    
    gps_mm : pandas.DataFrame
        DataFrame containing aggregated GPS data per trip, with columns 'traj', 'first_emission', 'last_emission', 'sub_edges'.
    
    dic_geohash : dict
        Dictionary mapping geohashes to integer indices.
    
    dic_candidates : dict
        Dictionary mapping geohashes to candidate edges.
    """
    gps.reset_index(drop=True, inplace=True)
    gps['id'] = range(len(gps))
    gps["id"] = gps["id"].astype(np.uint32)
    gps['geohash']   = pygeohash_fast.encode_many(gps['lon'].values, gps['lat'].values, 8)
    all_geohashes = set(gps['geohash'] )
    dic_geohash = dict(zip(all_geohashes, range(len(all_geohashes))))
    dic_candidates = {k: cand_edge[k] for k in cand_edge.keys() & all_geohashes}

    gps['edge'] = gps['geohash'].map(dic_candidates)
    gps['geohash_int'] = gps['geohash'].map(dic_geohash)
    gps.set_index('geohash_int', inplace=True)
    gps = gpd.GeoDataFrame(gps, geometry=gpd.points_from_xy(gps['lon'], gps['lat']), crs=4326)
    gps.to_crs(3035, inplace=True)
    gps['points'] = list(zip(gps['lon'], gps['lat']))
    
    gps_mm = gps.groupby('ID_trip').agg({'points' : lambda row: np.array(list(row), dtype=float),  
                                            "id":['first', 'last'], 
                                            "edge":  sum})
    gps_mm.columns = ['traj', 'first_emission', 'last_emission', 'sub_edges']
    gps_mm = gps_mm[gps_mm['sub_edges']!=0]
    gps_mm['sub_edges'] = gps_mm['sub_edges'].apply(lambda row: list(set(row)))
    gps.drop(columns=['geohash', 'points', "edge"], inplace=True)
    return(gps, gps_mm, dic_geohash, dic_candidates)


import os

import networkx as nx
import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix

from gpsmatcher.utils import load_param, print_step, save_param


def node2edge_graph(graph_nx, show_print=True):
    """
    Convert a node-weighted graph to an edge-weighted directed graph.

    Parameters
    ----------
    graph_nx : networkx.DiGraph
        The input graph with node weights.

    show_print : bool, optional
        Whether to print a message indicating the completion of the operation. Default is True.

    Returns
    -------
    new_graph : networkx.DiGraph
        Converted graph with edge weights.
    """
    edges = list(graph_nx.edges())
    edges2weight =nx.get_edge_attributes(graph_nx, 'weight')
    edges2id =nx.get_edge_attributes(graph_nx, 'edge_id')
    new_edges = []
    for from_node, to_node in edges:
        id_edge = edges2id[(from_node, to_node)]
        for i in graph_nx.in_edges(from_node):
            id_row = edges2id[i]
            weight = (edges2weight[(from_node, to_node)] + edges2weight[i]) / 2
            new_edges.append((id_row, id_edge, weight))
    new_graph = nx.DiGraph()
    new_graph.add_weighted_edges_from(new_edges)
    
    if show_print:
        print("Node to edge graph conversion done")
    return(new_graph)

def edge_graph_trans(new_graph, max_weight, show_print=True):
    """
    Compute edge-to-edge transitions in a weighted directed graph.

    Parameters
    ----------
    new_graph : networkx.DiGraph
        The input graph with edge weights.

    max_weight : float
        Maximum weight for considering transitions.

    show_print : bool, optional
        Whether to print a message indicating the completion of the operation. Default is True.

    Returns
    -------
    all_transition : pd.DataFrame
        DataFrame containing edge-to-edge transitions with columns 'ori_edge', 'dest_edge', 'weight'.
    """
    all_transition = []
    for id_node, ori_node in enumerate(list(new_graph.nodes)):
        one_edge_trans = nx.single_source_dijkstra_path_length(new_graph, ori_node, cutoff=max_weight)
        all_transition.append((ori_node, list(one_edge_trans.keys()), list(one_edge_trans.values())))
    all_transition = pd.DataFrame(all_transition, columns=['ori_edge', 'dest_edge', 'weight'])
    
    if show_print:
        print("Edge to edge transition computation done")
    return(all_transition)

def transition_matrix(graph_nx, max_weight, beta = 1/500, save=True, folder_name="mm_input", show_print=True):
    """
    Compute a transition matrix for a graph.

    Parameters
    ----------
    graph_nx : networkx.DiGraph
        The input graph.

    max_weight : float
        Maximum weight for considering transitions.

    beta : float, optional
        Scaling factor for the exponential term. Default is 1/500.

    save : bool, optional
        Whether to save the computed transition matrix. Default is True.

    folder_name : str, optional
        Folder name for saving/loading. Default is "mm_input".

    show_print : bool, optional
        Whether to print a message indicating the completion of the operation. Default is True.

    Returns
    -------
    transition_matrix : scipy.sparse.csr_matrix
        Computed transition matrix.
    """
    if show_print:
        print_step("Start process transition")
        
    if save and (os.path.isfile(folder_name + '/transition_matrix.pickle')):
        transition_matrix = load_param(folder_name, "transition_matrix", show_print=show_print)
        return(transition_matrix)
    else:
        new_graph = node2edge_graph(graph_nx, show_print=show_print)
        all_transition = edge_graph_trans(new_graph, max_weight, show_print=show_print)
        all_transition = all_transition.explode(['dest_edge', 'weight']).reset_index(drop=True).astype(np.int64)
        int_max_edge = max(all_transition[["ori_edge", "dest_edge"]].max()) + 1
        row = all_transition['ori_edge'].values
        col = all_transition['dest_edge'].values
        data = np.exp(- beta * all_transition['weight'].values)
        transition_matrix = csr_matrix((data, (row, col)), shape=(max(row)+1, max(col)+1))
        transition_matrix = (transition_matrix * 100).astype(np.int8)
        if save:
            save_param(folder_name, "transition_matrix", transition_matrix)
        return(transition_matrix)


import os
import pickle
from itertools import chain

import networkx as nx


def remove_unneccesarry_att(G, atts_to_keep = ['source', 'target', 'weight', 'geometry']):
    """
    Remove unnecessary attributes from edges in a graph.

    Parameters
    ----------
    G : networkx.DiGraph
        The input graph.

    atts_to_keep : list, optional
        List of attributes to keep. Default is ['source', 'target', 'weight', 'geometry'].

    Returns
    -------
    G : networkx.Graph
        The graph with only the specified attributes.
    """
    atts_name = list(set(chain.from_iterable(d.keys() for *_, d in G.edges(data=True))))
    atts_to_remove = [i for i in atts_name if i not in ['source', 'target', 'weight', 'geometry']]
    for n1, n2, d in G.edges(data=True):
        for att in atts_to_remove:
            d.pop(att, None)
    return(G)

def print_step(msg):
    """
    Print a step message with dashes above and below.

    Parameters
    ----------
    msg : str
        Message to print.
    """
    print("-"*len(msg))
    print(msg)
    print("-"*len(msg))
    return(None)

def load_param(folder_name, param_name, show_print=True):
    """
    Load a parameter from a pickle file.

    Parameters
    ----------
    folder_name : str
        Folder name containing the pickle file.

    param_name : str
        Name of the parameter.

    show_print : bool, optional
        Whether to print a message indicating the completion of the operation. Default is True.

    Returns
    -------
    param : object
        Loaded parameter.
    """
    with open('{}/{}.pickle'.format(folder_name, param_name), 'rb') as handle:
        param = pickle.load(handle)
    if show_print:
        print("Load existing {} done".format(param_name))
    return(param)

def save_param(folder_name, param_name, param):
    """
    Save a parameter to a pickle file.

    Parameters
    ----------
    folder_name : str
        Folder name for saving the pickle file.

    param_name : str
        Name of the parameter.

    param : object
        Parameter to save.
    """
    isExist = os.path.exists(folder_name)
    if not isExist:
        os.makedirs(folder_name)
    with open('{}/{}.pickle'.format(folder_name, param_name), 'wb') as handle:
        pickle.dump(param, handle, protocol=pickle.HIGHEST_PROTOCOL)
    return(None)
